/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

//created by Andrew Caldwell on 03/16/98  caldwell@cs.ucla.edu

// CHANGES

// 980320 ilm  safety measures in _set methods : a pointer can be newed
//                    only if it is NULL
// 980407 aec  added minCapacities data and functions
// 980520 ilm  added _computeXYTicSet(const vector<BBox>&);

#include <set.h>
#include "partProb.h"
#include "partitionData.h"
#include "u-linalg.h"
#include "termiProp.h"

PartitioningProblem::PartitioningProblem(
	const HGraphFixed& 		hgraph, 
	const PartitioningBuffer& 	solnBuffers,
	const Partitioning&		fixedConstr,
	const vector<BBox>&		partitions,
	const vector<vector<double> >&	capacities,
	const vector<double>&		tolerances,
	const vector<unsigned>& 	terminalToBlock,
	const vector<BBox>& 		padBlocks,
	Verbosity			verbosity)

	: _hgraph         (const_cast<HGraphFixed*>(&hgraph) ),
	  _solnBuffers    (const_cast<PartitioningBuffer*>(&solnBuffers)),
	  _fixedConstr     (const_cast<Partitioning*>(&fixedConstr) ),
	  _partitions     (const_cast<vector<BBox>*>(&partitions) ),
	  _capacities     (const_cast<vector<vector<double> >*>(&capacities) ),
          _maxCapacities  (NULL),
	  _minCapacities  (NULL),
          _totalWeight    (NULL),
	  _terminalToBlock(const_cast<vector<unsigned>*>(&terminalToBlock) ), 
	  _padBlocks      (const_cast<vector<BBox>*>(&padBlocks) ),
	  _verb(verbosity)
{
    abkfatal(_partitions->size() == _capacities->size(),
	"partition size does not match capacities size");
    abkfatal(tolerances.size() == (*_capacities)[0].size(),
	"tolerences size does not match capacitites.size");

    abkfatal(_terminalToBlock->size() >= _hgraph->getNumTerminals(),
	"terminalToBlock size does not match hgraph.numTerminals");

    _clusterDegrees = NULL;
    _ownsData = false;

    _setTotalWeight();
    _setMaxCapacities(tolerances, capacities);
    _setMinCapacities(tolerances, capacities);

    _postProcess();
}

void PartitioningProblem::setSolnBuffers(PartitioningBuffer* newbufs)
//  LOOK: deletes the old ones !!! will own the new ones !!! -ILM
//  to be used for reordering partitioning solutions
{
    abkfatal(_ownsData, 
		"can't call setSolnBuffers if PartProb does not own data")

    delete _solnBuffers;
    _solnBuffers=newbufs;
}

void PartitioningProblem::_freeDynamicData()
{
    if(_ownsData)
    {
        delete   _hgraph;
        if (_solnBuffers)
            delete _solnBuffers;

        delete   _fixedConstr;
        delete   _partitions;
        delete   _capacities;
        delete   _maxCapacities;
        delete   _minCapacities;
        delete   _totalWeight;
        delete   _partitionCenters;
        delete   _terminalToBlock;
        delete   _padBlocks;
        delete   _padBlockCenters;
        delete   _clusterDegrees;
    }
    else
    {       
        if(_partitionCenters != NULL)
            delete _partitionCenters;
        if(_padBlockCenters != NULL)
            delete _padBlockCenters;
	if(_maxCapacities != NULL)
	    delete _maxCapacities;
	if(_minCapacities != NULL)
	    delete _minCapacities;
	if(_totalWeight != NULL)
	    delete _totalWeight;
	if(_clusterDegrees != NULL)
	    delete _clusterDegrees;
    }
}

void PartitioningProblem::_postProcess()
{ 
    //_setTrivUSparseMatrix(); // does nothing if there is a USparseMat already
    if(_partitions != NULL)
	_setPartitionCenters();
    if(_padBlocks != NULL && _terminalToBlock != NULL)
	_setPadBlockCenters();
    _computeXYTics(*_partitions);

    _attributes.clearAll();

    if(_capacities != NULL)
	_attributes.hasCap = true;
    if(_maxCapacities != NULL)
	_attributes.hasMaxCap = true;
    if(_minCapacities != NULL)
	_attributes.hasMinCap = true;
    if(_partitions != NULL)
	_attributes.hasPartBBoxes = true;
    if(_partitionCenters != NULL)
	_attributes.hasPartCenters = true;
    if(_padBlocks != NULL)
	_attributes.hasPadBBoxes = true;
    if(_padBlockCenters != NULL)
	_attributes.hasPadCenters = true;
    if(_totalWeight != NULL)
	_attributes.hasTotalWeight = true;
    if(_clusterDegrees != NULL)
	_attributes.hasClusterDegrees = true;
}

void PartitioningProblem::_setTrivUSparseMatrix()
{
    if (_hgraph==NULL || _clusterDegrees!=NULL) return;
    _clusterDegrees
       =new USparseMatrix(_hgraph->getNumEdges(),_hgraph->getNumNodes());

    // fill the matrix here with 1s where appropriate
    for (itHGFNodeGlobal v=(*_hgraph).nodesBegin();
                        v!=(*_hgraph).nodesEnd(); ++v)
        for (itHGFEdgeLocal e=(*v)->edgesBegin(); e!=(*v)->edgesEnd(); e++)
            (*_clusterDegrees)((*e)->getIndex(),(*v)->getIndex())=1;
}


void PartitioningProblem::_setMaxCapacities(
		const vector<double>& tolerances,
		const vector<vector<double> >& capacities)
{
	//compute the max value for each partition/capacity
	// from tolerances and capacities

    if (_maxCapacities) return;

    unsigned numCapacities = tolerances.size();

    _maxCapacities = new vector<vector<double> >(capacities.size(),
					vector<double>(numCapacities));

    for(unsigned b = 0; b < capacities.size(); b++)
    {
	#ifdef PARTDEBUG
	    abkfatal(capacities[b].size() == numCapacities,
		"all partitions must have the same number of capacities");
	#endif

	copy(capacities[b].begin(), capacities[b].end(), 
		(*_maxCapacities)[b].begin());

        for(unsigned c = 0; c < numCapacities; c++)
	{
	    #ifdef PARTDEBUG
		abkfatal(tolerances[c] < 1.0, 
		  "suspicious tolerance > 1. Tolerances should be in percent");
	    #endif
	    (*_maxCapacities)[b][c] *= tolerances[c] + 1;
	}
    }
}

void PartitioningProblem::_setMinCapacities(
		const vector<double>& tolerances,
		const vector<vector<double> >& capacities)
{
	//compute the min value for each partition/capacity
	// from tolerances and capacities

    if (_minCapacities) return;

    unsigned numCapacities = tolerances.size();

    _minCapacities = new vector<vector<double> >(capacities.size(),
					vector<double>(numCapacities));

    for(unsigned b = 0; b < capacities.size(); b++)
    {
	#ifdef PARTDEBUG
	    abkfatal(capacities[b].size() == numCapacities,
		"all partitions must have the same number of capacities");
	#endif

	copy(capacities[b].begin(), capacities[b].end(), 
		(*_minCapacities)[b].begin());

        for(unsigned c = 0; c < numCapacities; c++)
	{
	    #ifdef PARTDEBUG
		abkfatal(tolerances[c] < 1.0, 
		  "suspicious tolerance > 1. Tolerances should be in percent");
	    #endif
	    (*_minCapacities)[b][c] *= 1 - tolerances[c];
	}
    }
}

void PartitioningProblem::_setPadBlockCenters()
{
    if (_padBlocks==NULL) return;
    _padBlockCenters = new vector<Point>(_padBlocks->size());
    for(unsigned k = 0; k < _padBlocks->size(); k++)
	(*_padBlockCenters)[k] = (*_padBlocks)[k].getGeomCenter();
}

void PartitioningProblem::_setPartitionCenters()
{
    _partitionCenters = new vector<Point>(_partitions->size());
    for(unsigned k = 0; k < _partitions->size(); k++)
	(*_partitionCenters)[k] = (*_partitions)[k].getGeomCenter();
}


void PartitioningProblem::_setTotalWeight()
{
	//compute the total non-terminal weight for each weight
    if (_totalWeight) return;
    abkfatal(_hgraph," Can't compute total weights --- no hypergraph"); 
    unsigned numWeights = _hgraph->getNumWeights();
    _totalWeight = new vector<double>(numWeights, 0.0);

    itHGFNodeGlobal nItr = _hgraph->nodesBegin()+_hgraph->getNumTerminals();
    for( ; nItr != _hgraph->nodesEnd(); nItr++)
    {
	for(unsigned w = 0; w < numWeights; w++)
	   (*_totalWeight)[w] += (*nItr)->getWeight(w);	    
    }
}

void PartitioningProblem::reserveBuffers(unsigned num)
{
    _solnBuffers->setEndUsedSoln(num);
}

// copy ctor
PartitioningProblem::PartitioningProblem(const PartitioningProblem &prob):
    _attributes(prob._attributes),_bestSolnNum(prob._bestSolnNum),
        _ownsData(true),_verb(prob._verb)
{
      _copyDynamicData(prob);
}

void PartitioningProblem::_copyDynamicData(const PartitioningProblem &prob)
{
    _capacities = new vector<vector<double> >(*prob._capacities);

    if (prob._clusterDegrees)
        _clusterDegrees = new USparseMatrix(*prob._clusterDegrees);
    else
        _clusterDegrees=NULL;

    _fixedConstr = new Partitioning(*prob._fixedConstr);
    _hgraph = new HGraphFixed(*prob._hgraph);

    if (prob._maxCapacities)
        _maxCapacities = new vector<vector<double> >(*prob._maxCapacities);
    else
        _maxCapacities=NULL;

    if (prob._minCapacities)
        _minCapacities = new vector<vector<double> >(*prob._minCapacities);
    else
        _minCapacities=NULL;

    if (prob._padBlockCenters)
        _padBlockCenters = new vector<Point>(*prob._padBlockCenters);
    else
        _padBlockCenters=NULL;

    _padBlocks = new vector<BBox>(*prob._padBlocks);

    if (prob._partitionCenters)
        _partitionCenters = new vector<Point>(*prob._partitionCenters);
    else
        _partitionCenters=NULL;

    _partitions = new vector<BBox>(*prob._partitions);

    _solnBuffers = new PartitioningBuffer(*prob._solnBuffers);

    _terminalToBlock = new vector<unsigned>(*prob._terminalToBlock);

    if (prob._totalWeight)
        _totalWeight = new vector<double>(*prob._totalWeight);
    else
        _totalWeight=NULL;
}


PartitioningProblem
&PartitioningProblem::operator=(const PartitioningProblem &prob)
{
    _attributes=prob._attributes;
    _bestSolnNum=prob._bestSolnNum;
    _verb=prob._verb;
    _freeDynamicData();
    _ownsData=true;
    _copyDynamicData(prob);
    return *this;
}

void PartitioningProblem::propagateTerminals(double fuzziness) // can be 10 (%)
{
    abkfatal(_terminalToBlock && _padBlocks," terminal block info missing ");
    PartitionIds freeToMove; 
    freeToMove.setToAll(_partitions->size());

    TerminalPropagator propag(*_partitions,fuzziness);
    vector<unsigned>::const_iterator tIt, cIt;
    unsigned termNo=0;
    for(tIt=_terminalToBlock->begin();tIt!=_terminalToBlock->end();tIt++)
    {
       PartitionIds& parts=(*_fixedConstr)[termNo];
       if (parts==freeToMove) parts=PartitionIds();
       propag.doOneTerminal((*_padBlocks)[*tIt],parts);
       termNo++;
    }

    termNo=0;
    for(unsigned k=0; k!= _solnBuffers->size(); k++)
    {
       Partitioning& soln= (*_solnBuffers)[k];
       for(cIt=_terminalToBlock->begin();cIt!=_terminalToBlock->end();cIt++)
       {
          PartitionIds& parts=soln[termNo];
          if (parts==freeToMove) parts=PartitionIds();
          propag.doOneTerminal((*_padBlocks)[*cIt],parts);
          termNo++;
       }
    }
    for(unsigned i=0; i!=_hgraph->getNumTerminals();i++)
    for(unsigned j=0; j!=_hgraph->getNumWeights();  j++)
        _hgraph->getNodeByIdx(i).setWeight(0.0,j);
//  _hgraph->clearTerminals();
}

void PartitioningProblem::_computeXYTics(const vector<BBox>& bboxes)
{
   set<double, less<double> > _xTicSet, _yTicSet;
   for(unsigned i=0; i!=bboxes.size(); i++)
   {
      _xTicSet.insert(bboxes[i].xMin);
      _xTicSet.insert(bboxes[i].xMax);
      _yTicSet.insert(bboxes[i].yMin);
      _yTicSet.insert(bboxes[i].yMax);
   }

/*
cout<<"num bboxes(partitions) "<<bboxes.size()<<endl;
cout<<"xTicSet: "<<_xTicSet.size()<<endl;
cout<<"yTicSet: "<<_yTicSet.size()<<endl;
cout<<"BOXES "<<bboxes<<endl;
*/

   if ( (_xTicSet.size()-1)*(_yTicSet.size()-1) != bboxes.size() )
   {
      if (_verb.forActions>1)
       abkwarn(0," BBoxes do not come from a grid (may be bloated/fuzzy)\n");
      return;
   }

   _xTics.clear();
   _yTics.clear();

   set<double, less<double> >::iterator setIt;
   for(setIt=_xTicSet.begin(); setIt!=_xTicSet.end(); setIt++)
       _xTics.push_back(*setIt);

   set<double, less<double> >::reverse_iterator rsetIt;
   for(rsetIt=_yTicSet.rbegin(); rsetIt!=_yTicSet.rend(); rsetIt++)
       _yTics.push_back(*rsetIt);

   abkfatal(_xTicSet.size()>1 && _yTicSet.size()>1,
            " Internal error in PartProblem: only one tic for x- and/or y- ");
}

extern double getBBoxQuantization
(unsigned quantizationBound, const vector<Point>& pts);

double PartitioningProblem::
getScalingFactorForBBoxQuantization(unsigned quantizationBound) const
{
  return  getBBoxQuantization(quantizationBound, *_partitionCenters);
}

extern double getBBoxQuantization
(unsigned quantizationBound, const vector<Point>& pts)
{

   abkfatal(pts.size(),"Can not process empty point-set");
   set<double, less<double> > _xTicSet, _yTicSet;
   for(unsigned i=0; i!=pts.size(); i++)
   {
      _xTicSet.insert(pts[i].x);
      _yTicSet.insert(pts[i].y);
   }

   if ( _xTicSet.size()*_yTicSet.size() != pts.size() )
      abkwarn(0," BBoxQuantization: Given points do not come from a grid.\n");

   vector<double> xcTics, ycTics;

   set<double, less<double> >::iterator setIt;
   for(setIt=_xTicSet.begin(); setIt!=_xTicSet.end(); setIt++)
       xcTics.push_back(*setIt);

   for(setIt=_yTicSet.begin(); setIt!=_yTicSet.end(); setIt++)
       ycTics.push_back(*setIt);

   int k;
   for(k=xcTics.size()-1; k>=0; k--) xcTics[k] -= xcTics[0]; 
   for(k=ycTics.size()-1; k>=0; k--) ycTics[k] -= ycTics[0];

// this is somewhat redundant for now, but will be used later

   abkfatal(xcTics.size()>0 && ycTics.size()>0, "Internal error in PartProb");

   if (xcTics.size()==1)
   {
      if (ycTics.size()==1) return 1.0;
      return 1/ycTics[1];
   }
   else if (ycTics.size()==1) return 1/xcTics[1];

   double delta = 1.0 / (max(xcTics.back(), ycTics.back()));
   double a     = quantizationBound / (xcTics.back() + ycTics.back());
   double dx=xcTics[1];
   double dy=ycTics[1];

// Now we will assume that finding a delta-gauge of dx and dy is enough
// (improve later)

   double result = a / abkGcd(rint(dx * a), rint(dy * a));

   if (result<delta) result=delta;

   abkfatal3(result*(xcTics.back() + ycTics.back()) < quantizationBound,
            " BBox quantization failed: quantization bound ",quantizationBound,
            " exceeded ");

   return result;
}

